# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from git import Repo

from zubbi.scraper.connections.git import GitConnection
from zubbi.scraper.exceptions import CheckoutError
from zubbi.scraper.repos.git import FileContent, GitRepository


@pytest.mark.parametrize(
    "url, repo_name, expected",
    [
        ("http://localhost", "foo/bar", "http://localhost/foo/bar"),
        ("https://localhost/git", "foo.bar", "https://localhost/git/foo.bar"),
    ],
)
def test_get_remote_url(url, repo_name, expected):
    git_con = GitConnection(url)
    remote_url = git_con.get_remote_url(repo_name)
    assert remote_url == expected


def test_get_repo_object_existing_path(mock_git_repo, tmpdir):
    git_url = "https://localhost/git"
    repo_name = "foo"

    with mock_git_repo(tmpdir, repo_name, git_url):
        git_con = GitConnection(git_url, workspace=tmpdir)
        # The __init__() method calls _get_repo_object()
        git_repo = GitRepository(repo_name, git_con)
        assert isinstance(git_repo._repo, Repo)


def test_get_repo_object_failure(tmpdir):
    git_url = "https://localhost/git"
    repo_name = "foo"

    git_con = GitConnection(git_url, workspace=tmpdir)
    git_repo = GitRepository(repo_name, git_con)
    assert git_repo._repo is None


def test_check_out_file(mock_git_repo, tmpdir):
    git_url = "https://localhost/git"
    repo_name = "foo"

    with mock_git_repo(tmpdir, repo_name, git_url):
        git_con = GitConnection(git_url, workspace=tmpdir)
        git_repo = GitRepository(repo_name, git_con)

        contents = git_repo.check_out_file("README")
        assert contents == "Repository: foo"


def test_check_out_non_existing_file(mock_git_repo, tmpdir):
    git_url = "https://localhost/git"
    repo_name = "foo"

    with mock_git_repo(tmpdir, repo_name, git_url):
        git_con = GitConnection(git_url, workspace=tmpdir)
        git_repo = GitRepository(repo_name, git_con)

        with pytest.raises(CheckoutError) as excinfo:
            git_repo.check_out_file("non-existing-file")
        assert "Failed to check out 'non-existing-file'" in str(excinfo.value)


def test_list_directory(mock_git_repo, tmpdir):
    git_url = "https://localhost/git"
    repo_name = "foo"

    with mock_git_repo(tmpdir, repo_name, git_url):
        git_con = GitConnection(git_url, workspace=tmpdir)
        git_repo = GitRepository(repo_name, git_con)

        contents = git_repo.list_directory("/")
        # The contents dict should contain an entry for the README file
        readme_contents = contents["README"]
        assert isinstance(readme_contents, FileContent)
        assert readme_contents.path == "README"


def test_list_non_existing_directory(mock_git_repo, tmpdir):
    git_url = "https://localhost/git"
    repo_name = "foo"

    with mock_git_repo(tmpdir, repo_name, git_url):
        git_con = GitConnection(git_url, workspace=tmpdir)
        git_repo = GitRepository(repo_name, git_con)

        with pytest.raises(CheckoutError) as excinfo:
            git_repo.list_directory("/non-existing-directory")
        assert "Failed to check out '/non-existing-directory/'" in str(excinfo.value)



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timezone

import arrow
import pkg_resources


def urljoin(base, *parts):
    clean_parts = (p.strip("/") for p in parts)
    return "/".join((base.rstrip("/"), *clean_parts))


def last_changed_from_blame_range(start, end, blames):
    if not blames:
        return
    # Sort the blames by date (newest first)
    blames = sorted(blames, key=lambda b: b["date"], reverse=True)
    # Build a range including each line (+1 to ensure the last line is also
    # part of the range)
    line_range = set(range(start, end + 1))
    # Try to find the first blame (newest one) that is matching the line range
    for blame in blames:
        blame_range = range(blame["start"], blame["end"] + 1)
        # Find intersections between both ranges
        # If we found one, we return the date (as the first match is also the
        # newest one)
        if line_range.intersection(blame_range):
            return blame["date"]


def prettydate(date):
    now = datetime.now(timezone.utc)
    """
    Return the relative timeframe between the given date and now.

    e.g. 'Just now', 'x days ago', 'x hours ago', ...

    When the difference is greater than 7 days, the timestamp will be returned
    instead.
    """
    diff = now - date
    # Show the timestamp rather than the relative timeframe when the difference
    # is greater than 7 days
    if diff.days > 7:
        return date.strftime("%d. %b %Y")
    return arrow.get(date).humanize()


def get_version():
    try:
        version = pkg_resources.get_distribution("zubbi").version
    except pkg_resources.DistributionNotFound:
        version = "unknown"
    return version



from unittest import mock

import elasticsearch_dsl
import pytest
from elasticsearch_dsl.serializer import serializer

from zubbi.models import (
    AnsibleRole,
    GitRepo,
    init_elasticsearch_con,
    ZuulJob,
    ZuulTenant,
)


@pytest.fixture(scope="function")
def elmock():
    with mock.patch.object(elasticsearch_dsl.connections, "Elasticsearch") as _elmock:
        yield _elmock

    # Reset the name of each index to its original value as it might have changed
    # during the tests. Btw, this clearly shows that this is a hack ;-)
    for idx_cls in [ZuulJob, AnsibleRole, ZuulTenant, GitRepo]:
        # As the Index.name attribute holds the constant value from the original
        # definition, we can use it to reset the active value which might have
        # changed due to the es_index_prefix hack.
        idx_cls._index._name = idx_cls.Index.name


def test_elasticsearch_init(elmock):

    # Define the existing indices for the mock
    existing_indices = {"zuul-jobs", "ansible-roles", "unknown-index"}
    elmock.return_value.indices.exists.side_effect = (
        lambda index: index in existing_indices
    )

    init_elasticsearch_con("127.0.0.1", "user", "password")

    # Validate that the Elasticsearch() (which is called by elasticsearch-dsl in the end)
    # was called with the correct arguments.
    assert elmock.call_args == mock.call(
        host="127.0.0.1",
        port=9200,
        http_auth=("user", "password"),
        serializer=serializer,
    )

    # Validate that all necessary indices were checked for existence
    checked_indices = {
        call[1]["index"] for call in elmock.return_value.indices.exists.call_args_list
    }
    assert {
        "zuul-jobs",
        "ansible-roles",
        "zuul-tenants",
        "git-repos",
    } == checked_indices

    # Validate that only the missing indices were created
    created_indices = {
        call[1]["index"] for call in elmock.return_value.indices.create.call_args_list
    }
    assert {"zuul-tenants", "git-repos"} == created_indices


def test_elasticsearch_init_with_prefix(elmock):

    # Define the existing indices for the mock
    existing_indices = {"zubbi-zuul-jobs", "zubbi-ansible-roles", "unknown-index"}
    elmock.return_value.indices.exists.side_effect = (
        lambda index: index in existing_indices
    )

    init_elasticsearch_con("127.0.0.1", "user", "password", es_index_prefix="zubbi")

    # Validate that the Elasticsearch() (which is called by elasticsearch-dsl in the end)
    # was called with the correct arguments.
    assert elmock.call_args == mock.call(
        host="127.0.0.1",
        port=9200,
        http_auth=("user", "password"),
        serializer=serializer,
    )

    # Validate that all necessary indices were checked for existence
    checked_indices = {
        call[1]["index"] for call in elmock.return_value.indices.exists.call_args_list
    }
    assert {
        "zubbi-zuul-jobs",
        "zubbi-ansible-roles",
        "zubbi-zuul-tenants",
        "zubbi-git-repos",
    } == checked_indices

    # Validate that only the missing indices were created
    created_indices = {
        call[1]["index"] for call in elmock.return_value.indices.create.call_args_list
    }
    assert {"zubbi-zuul-tenants", "zubbi-git-repos"} == created_indices


def test_elasticsearch_init_with_prefix_multi(elmock):

    # Define the existing indices for the mock
    existing_indices = {"zubbi-zuul-jobs", "zubbi-ansible-roles", "unknown-index"}
    elmock.return_value.indices.exists.side_effect = (
        lambda index: index in existing_indices
    )

    init_elasticsearch_con("127.0.0.1", "user", "password", es_index_prefix="zubbi")
    init_elasticsearch_con("127.0.0.1", "user", "password", es_index_prefix="zubbi")

    # Evenv if we called the init() method multiple times, the index should only be
    # prepended once.
    checked_indices = {
        call[1]["index"] for call in elmock.return_value.indices.exists.call_args_list
    }
    assert {
        "zubbi-zuul-jobs",
        "zubbi-ansible-roles",
        "zubbi-zuul-tenants",
        "zubbi-git-repos",
    } == checked_indices


def test_elasticsearch_write(elmock):
    init_elasticsearch_con("127.0.0.1", "user", "password")

    zt = ZuulTenant(name="foo")
    zt.save()

    assert elmock.return_value.index.call_args == mock.call(
        doc_type="doc", index="zuul-tenants", body=zt.to_dict()
    )


def test_elasticsearch_write_with_prefix(elmock):
    init_elasticsearch_con("127.0.0.1", "user", "password", es_index_prefix="zubbi")

    zt = ZuulTenant(name="foo")
    zt.save()

    assert elmock.return_value.index.call_args == mock.call(
        doc_type="doc", index="zubbi-zuul-tenants", body=zt.to_dict()
    )



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from zubbi.scraper.connections.gerrit import CGitUrlBuilder, GitwebUrlBuilder


@pytest.mark.parametrize(
    "gerrit_url, repo_name, file_path, highlight_start, expected",
    [
        (
            "https://gerrit.example.com/",
            "bar",
            "roles/bar/README.md",
            1,
            "https://gerrit.example.com/bar/tree/roles/bar/README.md#n1",
        ),
        (
            "git.example.org/cgit",
            "foo",
            "config.yaml",
            None,
            "git.example.org/cgit/foo/tree/config.yaml",
        ),
    ],
)
def test_url_for_file_cgit(gerrit_url, repo_name, file_path, highlight_start, expected):
    url_builder = CGitUrlBuilder(gerrit_url)
    file_url = url_builder.build_file_url(repo_name, file_path, highlight_start, None)
    assert file_url == expected


@pytest.mark.parametrize(
    "gerrit_url, repo_name, file_path, highlight_start, expected",
    [
        (
            "https://gerrit.example.com/",
            "bar",
            "roles/bar/README.md",
            1,
            "https://gerrit.example.com/?p=bar.git;a=blob;f=roles/bar/README.md#l1",
        ),
        (
            "git.example.org/gitweb",
            "foo",
            "config.yaml",
            None,
            "git.example.org/gitweb?p=foo.git;a=blob;f=config.yaml",
        ),
    ],
)
def test_url_for_file_gitweb(
    gerrit_url, repo_name, file_path, highlight_start, expected
):
    url_builder = GitwebUrlBuilder(gerrit_url)
    file_url = url_builder.build_file_url(repo_name, file_path, highlight_start, None)
    assert file_url == expected



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import render_template, request

from .utils import prettydate


def make_error_handler(template="error.html"):
    def _handler(e):
        return (
            render_template(template, error_code=e.code, error_message=e.description),
            e.code,
        )

    return _handler


def current_endpoint(endpoint):
    blueprint_name = request.blueprint
    # Logic from flask.helpers.url_for
    if endpoint[:1] == ".":
        if blueprint_name is not None:
            endpoint = blueprint_name + endpoint
        else:
            endpoint = endpoint[1:]
    return request.endpoint == endpoint


def init_helpers(app):
    app.add_template_test(current_endpoint)
    app.add_template_filter(prettydate)


def calculate_pagination(current_page, last_page, step_size=2):
    start = 0
    end = last_page
    steps = []

    if current_page + step_size < last_page - 1:
        # We have a gap on the end
        end = current_page + step_size

    if current_page > step_size + 1:
        # We have a gap on the start
        start = current_page - step_size
        steps.append([0])

    steps.append([i for i in range(start, end + 1)])

    if end != last_page:
        steps.append([last_page])

    pagination = {"current": current_page, "last_page": last_page, "steps": steps}
    return pagination



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import hashlib
import hmac
import json
import math

from elasticsearch_dsl import Q
from flask import (
    abort,
    current_app,
    jsonify,
    make_response,
    redirect,
    render_template,
    request,
    url_for,
)
from flask.views import MethodView

from .extensions import get_zmq_socket
from .helpers import calculate_pagination
from .models import block_type, BlockSearch, class_from_block_type
from .utils import get_version

SEARCHABLE_FIELDS = frozenset(["name", "description", "tenants", "repo"])
DEFAULT_SEARCH_FIELDS = frozenset(["name", "description"])
SEARCHABLE_BLOCK_TYPES = frozenset(["job", "role"])

WEBHOOK_EVENTS = ["installation", "installation_repositories", "push"]


def json_abort(status, msg=None):
    response = {"error": status}
    if msg is not None:
        response["msg"] = msg
    abort(make_response(jsonify(response), status))


class ZubbiMethodView(MethodView):
    @property
    @abc.abstractmethod
    def endpoint(self):
        pass

    @property
    @abc.abstractmethod
    def rule(self):
        pass

    @property
    @abc.abstractmethod
    def template_name(self):
        pass

    @classmethod
    def register_url(cls, app, **options):
        app.add_url_rule(cls.rule, view_func=cls.as_view(cls.endpoint), **options)

    def get_context(self, **kwargs):
        # Initialize context with meta fields that should be available on all pages
        context = {"meta": {"version": get_version()}}

        # Add additionally provided kwargs
        context = {**context, **kwargs}
        return context


class IndexView(ZubbiMethodView):

    endpoint = "index"
    rule = "/"
    template_name = "index.html"

    def get(self):
        context = self.get_context()
        return render_template(self.template_name, **context)


class SearchView(ZubbiMethodView):

    endpoint = "search"
    rule = "/search"
    template_name = "search.html"

    def get(self):
        query = request.args.get("query")
        advanced = request.args.get("advanced")
        if query is None:
            context = {
                "result": None,
                "advanced": advanced,
                "default_search_fields": DEFAULT_SEARCH_FIELDS,
                "searchable_block_types": SEARCHABLE_BLOCK_TYPES,
                "searchable_fields": SEARCHABLE_FIELDS,
            }
            context = {**self.get_context(), **context}
            return render_template(self.template_name, **context)

        exact = request.args.get("exact")
        block_filter = request.args.getlist("block_filter")
        fields = request.args.getlist("fields")

        try:
            if len(block_filter) == 1:
                block_class = class_from_block_type(block_filter[0].lower())
            else:
                # Since we only have two block types
                # the filter doesn't make sense in this case
                block_class = None
            filter_from = int(request.args.get("from", 0))
            filter_size = min(
                int(request.args.get("size", current_app.config["SEARCH_BATCH_SIZE"])),
                current_app.config["SEARCH_BATCH_LIMIT"],
            )
            if fields:
                field_set = set(fields)
                # Validate fields
                if not field_set.issubset(SEARCHABLE_FIELDS):
                    abort(400, "The given field is not searchable")

            else:
                field_set = set(DEFAULT_SEARCH_FIELDS)

            # Expand 'name' field to match jobs and roles
            if "name" in field_set:
                field_set.remove("name")
                field_set.update(["job_name", "role_name"])

        except ValueError:
            abort(400)

        filter_to = filter_size + filter_from

        # TODO: This could later be extended with:
        # extra_filter = [Q('term', tenants='<tenant>') | Q('term', private=False)]
        # if we have the tenant information via an Apache Mellon header.
        extra_filter = Q("term", private=False)
        search = BlockSearch(block_class=block_class).search_query(
            query, field_set, exact, extra_filter
        )
        result = search[filter_from:filter_to].execute()

        if len(result) == 1 and filter_from == 0:
            block = result[0]
            return redirect(
                url_for(
                    "zubbi.details",
                    repo=block.repo,
                    block_type=block_type(block),
                    name=block.name,
                ),
                code=303,
            )

        current_page = filter_from // filter_size
        # The last results fit on the second last page. E.g. If we have 9/18/... results
        last_page = max(math.ceil(result.hits.total / filter_size) - 1, 0)
        pagination = calculate_pagination(current_page, last_page)

        context = {
            "result": result,
            "query": query,
            "exact": exact,
            "fields": fields,
            "block_filter": block_filter,
            "filter_from": filter_from,
            "pagination": pagination,
            "batch_size": filter_size,
            "advanced": advanced,
            "default_search_fields": DEFAULT_SEARCH_FIELDS,
            "searchable_block_types": SEARCHABLE_BLOCK_TYPES,
            "searchable_fields": SEARCHABLE_FIELDS,
        }
        context = {**self.get_context(), **context}
        return render_template(self.template_name, **context)


class DetailView(ZubbiMethodView):

    endpoint = "details"
    rule = "/detail/<path:repo>/<block_type>/<name>"
    template_name = "details.html"

    def get(self, repo, block_type, name):
        try:
            BlockClass = class_from_block_type(block_type)
        except ValueError:
            abort(400, "Unknown block type '{}'".format(block_type))

        extra_filter = Q("term", private=False)
        search = BlockSearch(block_class=BlockClass).detail_query(
            name, repo, extra_filter
        )
        result = search.execute()

        if not result:
            abort(404, "No {} found with the name '{}'".format(block_type, name))

        context = self.get_context(block_type=block_type, block=result[0])
        return render_template(self.template_name, **context)


class HowToView(ZubbiMethodView):

    endpoint = "how-to"
    rule = "/how-to"
    template_name = "how-to.html"

    def get(self):
        context = self.get_context()
        return render_template(self.template_name, **context)


class AutoCompleteView(ZubbiMethodView):

    endpoint = "auto-complete"
    rule = "/api/search/autocomplete"
    template_name = None

    def get(self):
        term = request.args.get("term")

        if not term:
            json_abort(404)

        # TODO: This could later be extended to filter also for tenants
        # contexts = {
        #     "tenants": ["<tenant>"],
        #     "private": [False]
        # }
        # if we have the tenant information via an Apache Mellon header.
        contexts = {"private": [False]}
        search = BlockSearch().suggest_query(term, size=10, contexts=contexts)
        result = search.execute()

        json_response = set()

        for suggester in result.suggest.suggester:
            for option in suggester.options:
                # NOTE (fschmidt): If we need the result as block class (AnsibleRole,
                # ZuulJob), we could activate the following line
                # block = search._get_result(option.to_dict())
                # json_response.append(block.name)
                json_response.add(option.text)

        return jsonify(list(json_response))


class WebhookView(ZubbiMethodView):

    endpoint = "webhook"
    rule = "/api/webhook"
    template_name = None

    def post(self):
        # Get the payload in the expected format (json)
        # This should also catch payloads which are not json
        payload = request.get_json()
        if not payload:
            json_abort(400, "Payload is missing or not a valid JSON")

        # Check if we received a webhook from GitHub
        if request.headers.get("x-github-delivery") is None:
            json_abort(400, "X-GitHub-Delivery header missing.")

        # Verify that the webhook was sent from our own GitHub app
        # NOTE (fschmidt): We need the raw payload to make the verification work
        self.verify_signature(request.get_data(), request.headers)

        event = self.check_event(request.headers)

        if event is not None:
            get_zmq_socket().send_multipart(
                (event.encode("utf-8"), json.dumps(payload).encode("utf-8"))
            )

        return jsonify({"event_processed": bool(event)})

    def verify_signature(self, payload, headers):
        """Verify that the payload was sent from our GitHub instance."""
        github_signature = headers.get("x-hub-signature")
        if not github_signature:
            json_abort(401, "X-Hub-Signature header missing.")

        gh_webhook_secret = current_app.config["GITHUB_WEBHOOK_SECRET"]
        signature = "sha1={}".format(
            hmac.new(
                gh_webhook_secret.encode("utf-8"), payload, hashlib.sha1
            ).hexdigest()
        )

        if not hmac.compare_digest(signature, github_signature):
            json_abort(401, "Request signature does not match calculated signature.")

    def check_event(self, headers):
        # For now, we are only interested in the following events:
        # - installation {created, deleted}
        # - installation_repositories {added, removed}
        # - push
        event = headers.get("x-github-event")

        if event not in WEBHOOK_EVENTS:
            return None

        return event


class LmzifyView(IndexView):

    endpoint = "lmzify"
    rule = "/lmzify"
    template_name = "lmzify.html"

    def get(self):
        term = request.args.get("term")
        context = self.get_context(term=term)
        return render_template(self.template_name, **context)



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
from pathlib import Path

import pytest
from git import Repo


@pytest.fixture(scope="function")
def mock_git_repo():
    # NOTE (felix): I found that this is better put into a fixture, but I wanted
    # to have the workspace, repo_name and url configurable.
    @contextlib.contextmanager
    def _repo(workspace, repo_name, url):
        repo_workspace = Path(workspace) / repo_name
        mocked_repo = Repo.init(repo_workspace)
        mocked_repo.create_remote("origin", url)

        # Create a readme file and commit it to master
        readme = repo_workspace / "README"
        readme.write_text("Repository: {}".format(repo_name))
        mocked_repo.index.add([str(readme)])
        mocked_repo.index.commit("Initial commit")

        yield mocked_repo

        # TODO (felix): Usually we should do a cleanup here, but as this function is
        # currently only called with a tmpdir, we can do that later on.

    return _repo



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock

import pytest

import zubbi
from zubbi.scraper.connections.gerrit import CGitUrlBuilder, GerritConnection
from zubbi.scraper.connections.git import GitConnection
from zubbi.scraper.connections.github import GitHubConnection
from zubbi.scraper.exceptions import ScraperConfigurationError
from zubbi.scraper.main import init_connections


@pytest.fixture(scope="function")
def patch_es(monkeypatch):
    # Patch the function where it is used, not the original one
    monkeypatch.setattr(zubbi.scraper.main, "init_elasticsearch_con", mock.Mock())


def test_init_gerrit_con(patch_es):
    config = {
        "ES_HOST": "localhost",
        "CONNECTIONS": {
            "gerrit_con": {
                "provider": "gerrit",
                "url": "https://localhost/gerrit",
                "web_url": "https://localhost/gerrit-web",
                "user": "spam",
                "password": "eggs",
            }
        },
    }

    expected_con_data = {
        "git_host_url": "https://localhost/gerrit",
        "user": "spam",
        "password": "eggs",
        "workspace_dir": "/tmp/zubbi_working_dir",
        "base_url": "https://localhost/gerrit",
        "gitweb_url": "https://localhost/gerrit-web",
        "gitweb_type": "cgit",
    }

    connections = init_connections(config)
    gerrit_con = connections["gerrit_con"]

    assert isinstance(gerrit_con, GerritConnection)
    assert isinstance(gerrit_con.web_url_builder, CGitUrlBuilder)

    con_data = vars(gerrit_con)
    # That one is already checked via isinstance
    con_data.pop("web_url_builder")
    assert con_data == expected_con_data


def test_init_gerrit_con_invalid_webtype(patch_es):
    config = {
        "ES_HOST": "localhost",
        "CONNECTIONS": {
            "gerrit_con": {
                "provider": "gerrit",
                "url": "https://localhost/gerrit",
                "web_url": "https://localhost/gerrit-web",
                "web_type": "INVALID",
                "user": "spam",
                "password": "eggs",
            }
        },
    }

    with pytest.raises(ScraperConfigurationError) as excinfo:
        init_connections(config)
    assert "unsupported web_type 'INVALID'" in str(excinfo)


def test_init_github_con(patch_es, mock_github_api_endpoints):

    mock_github_api_endpoints("https://localhost/github")
    config = {
        "ES_HOST": "localhost",
        "CONNECTIONS": {
            "github_con": {
                "provider": "github",
                "url": "https://localhost/github",
                "app_id": 3,
                "app_key": os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "../../testdata/app_key_file",
                ),
            }
        },
    }

    expected_con_data = {
        "base_url": "https://localhost/github",
        "api_url": "https://localhost/github/api/v3",
        "graphql_url": "https://localhost/github/api/graphql",
        "app_id": 3,
    }

    connections = init_connections(config)
    github_con = connections["github_con"]
    assert isinstance(github_con, GitHubConnection)
    # Check that the expected data is a subset of the connection's underlying dict
    for key, val in expected_con_data.items():
        assert val == github_con.__dict__[key]


def test_init_git_con(patch_es):
    config = {
        "ES_HOST": "localhost",
        "CONNECTIONS": {
            "git_con": {
                "provider": "git",
                "url": "https://localhost/git",
                # Optional, if authentication is required
                "user": "foo",
                "password": "bar",
            }
        },
    }

    expected_con_data = {
        "git_host_url": "https://localhost/git",
        "user": "foo",
        "password": "bar",
        "workspace_dir": "/tmp/zubbi_working_dir",
    }

    connections = init_connections(config)
    git_con = connections["git_con"]

    assert isinstance(git_con, GitConnection)
    assert expected_con_data == git_con.__dict__


def test_init_con_invalid_provider(patch_es):
    config = {
        "ES_HOST": "localhost",
        "CONNECTIONS": {"invalid_con": {"provider": "invalid"}},
    }

    with pytest.raises(ScraperConfigurationError) as excinfo:
        init_connections(config)
    assert (
        "Could not init connection 'invalid_con'. Specified provider 'invalid' is not available"
        in str(excinfo)
    )



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from pathlib import Path

from git import Repo
from git.exc import GitCommandError, InvalidGitRepositoryError

from zubbi.scraper.exceptions import CheckoutError
from zubbi.scraper.repos import Repository


LOGGER = logging.getLogger(__name__)

# TODO Currently, there is no way to specify a different default branch for a "generic" git repository.
# If Gerrit (which will use this implementation) has some similar concept of
# default branches, we have to find a way to implement this.
DEFAULT_BRANCH = "master"


class GitRepository(Repository):
    def __init__(self, repo_name, git_con):
        self.repo_name = repo_name
        self.workspace_dir = Path(git_con.workspace_dir)
        # Build the remote url based on the gerrit connection parameters
        self.remote_url = git_con.get_remote_url(repo_name)
        self._repo = self._get_repo_object(retry=True)

    def _get_repo_object(self, retry=False):
        # Clone the repository if it does not exist, otherwise just fetch it
        # and reset the HEAD.
        repo_src_path = self.workspace_dir / self.repo_name
        repo = None
        if repo_src_path.exists():
            try:
                repo = Repo(repo_src_path)
                # TODO fetch and reset HEAD
                # TODO Which remote?
                repo.remotes["origin"].fetch(DEFAULT_BRANCH)
            except GitCommandError as e:
                LOGGER.error("Fetching repo '%s' failed: %s" % (self.repo_name, e))
            except InvalidGitRepositoryError as e:
                LOGGER.error(
                    "Could not use existing repository in '%s': %s" % (repo_src_path, e)
                )
        else:
            try:
                repo = Repo.clone_from(
                    self.remote_url, repo_src_path, bare=True, depth=1
                )
            except GitCommandError as e:
                LOGGER.error("Cloning repo '%s' failed: %s" % (self.repo_name, e))

        if repo is None and retry:
            LOGGER.info("Retrying clone/fetch once")
            return self._get_repo_object()

        return repo

    def check_out_file(self, file_path):
        try:
            LOGGER.debug("Checking out '%s'", file_path)
            content = self._repo.git.show("{}:{}".format(DEFAULT_BRANCH, file_path))
            return content
        except GitCommandError as e:
            raise CheckoutError(file_path, e.stderr)

    def list_directory(self, directory_path):
        LOGGER.debug("Listing contents of '%s' directory", directory_path)
        command = ["git", "ls-tree", "--name-only", DEFAULT_BRANCH]
        # git ls-tree uses the root of the repository automatically, if no path is provided
        # If we provide '/' instead, it will fail.
        if directory_path != "/":
            # We must ensure that the path ends with a slash if we want to get
            # the directory contents
            if not directory_path.endswith("/"):
                directory_path = "{}/".format(directory_path)
            command.append(directory_path)

        try:
            files = self._repo.git.execute(command).split()
            # To be compatible with the current GitHub implementation, the resulting
            # dictionary must provide the filename as key and a Contents-like object
            # as value.
            return {Path(f).name: FileContent(f) for f in files}
        except GitCommandError as e:
            raise CheckoutError(directory_path, e.stderr)

    def last_changed(self, path):
        # TODO Implement...
        pass

    def blame(self, path):
        # TODO Implement...
        pass

    def url_for_file(self, file_path, highlight_start=None, highlight_end=None):
        # NOTE (fschmidt): This does not make sense for plain git repositories.
        # It should be implemented by more concrete provider implementations like
        # GitHub or Gerrit.
        return None

    def url_for_directory(self, directory_path):
        # NOTE (fschmidt): This does not make sense for plain git repositories.
        # It should be implemented by more concrete provider implementations like
        # GitHub or Gerrit.
        return None

    @property
    def url(self):
        # NOTE (fschmidt): This does not make sense for plain git repositories.
        # It should be implemented by more concrete provider implementations like
        # GitHub or Gerrit.
        return None

    @property
    def private(self):
        # NOTE (fschmidt): This does not make sense for plain git repositories.
        # It should be implemented by more concrete provider implementations like
        # GitHub or Gerrit.
        return False

    @property
    def name(self):
        return self.repo_name


class FileContent:
    """Minimalistic class that provides the same API as GitHub's Contents class."""

    def __init__(self, path):
        self.path = path



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from zubbi.doc import render_markdown, render_sphinx


@pytest.mark.parametrize(
    "readme, expected",
    [
        ("Hello World!", "<p>Hello World!</p>\n"),
        ("**Hello World!**", "<p><strong>Hello World!</strong></p>\n"),
        ("# Hello World!", "<h1>Hello World!</h1>\n"),
    ],
)
def test_render_markdown_simple(readme, expected):
    result = render_markdown(readme)
    assert expected == result["html"]


@pytest.mark.parametrize(
    "readme, expected",
    [
        ("Hello World!", "<p>Hello World!</p>\n"),
        ("**Hello World!**", "<p><strong>Hello World!</strong></p>\n"),
    ],
)
def test_render_sphinx_simple(readme, expected):
    result = render_sphinx(readme)
    assert expected == result["html"]


def test_render_sphinx_python_code(readme_python_code):
    expected_code = (
        '<div class="highlight-python notranslate">'
        '<div class="highlight"><pre><span></span>'
        '<span class="kn">import</span> <span class="nn">'
        "this</span>\n</pre></div>\n</div>\n"
    )

    result = render_sphinx(readme_python_code)
    assert expected_code == result["html"]


def test_render_sphinx_yaml_code(readme_yaml_code):
    expected_code = (
        '<div class="highlight-yaml notranslate"><div class="highlight">'
        '<pre><span></span><span class="p p-Indicator">-</span>'
        ' <span class="nt">job</span><span class="p">:</span>\n'
        '    <span class="nt">name</span><span class="p">:</span>'
        ' <span class="l l-Scalar l-Scalar-Plain">foo</span>\n'
        '    <span class="nt">parent</span><span class="p">:</span>'
        ' <span class="l l-Scalar l-Scalar-Plain">bar</span>\n'
        '    <span class="nt">playbook</span><span class="p">:</span>'
        ' <span class="l l-Scalar l-Scalar-Plain">foo-bar.yaml</span>\n'
        "</pre></div>\n"
        "</div>\n"
    )

    result = render_sphinx(readme_yaml_code)
    assert expected_code == result["html"]


def test_render_sphinx_supported_os(readme_supported_os):
    expected_platforms = ["linux", "windows"]
    expected_html = "<p>This works on Linux and Windows!</p>\n"

    result = render_sphinx(readme_supported_os)

    assert expected_platforms == result["platforms"]
    assert expected_html == result["html"]



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timezone

from zubbi.scraper.repo_parser import RepoParser


JOB_1_SHA = "83bf1474a8a84cc1dddc8da435e2e4d9f4bbdeb1"
JOB_2_SHA = "1e4008e1c0f4511579e49854c34905b419441331"
JOB_3_SHA = "728e71a5ab9d095d7b983aee6db6ea4656072658"
ROLE_1_SHA = "a336b3fdba7cd8ad535350fd7d249b7285670dfd"
ROLE_2_SHA = "f39ee170837cb6a91cbb3d7f9def7584a3a27919"


def test_parse(repo_data):
    scrape_time = datetime.now(timezone.utc)
    # Extract test data from fixture function result
    repo, tenants, job_files, role_files = repo_data

    jobs, roles = RepoParser(repo, tenants, job_files, role_files, scrape_time).parse()

    # We assume that we can access the resulting jobs and roles dictionary
    # with the given SHA values. Otherwise, we will get a KeyError.
    job_1 = jobs[0]
    job_2 = jobs[1]
    job_3 = jobs[2]
    role_1 = [r for r in roles if r["role_name"] == "foo"][0]
    role_2 = [r for r in roles if r["role_name"] == "bar"][0]

    expected_job_1 = {
        "job_name": "my-cool-new-job",
        "repo": "my/project",
        "tenants": ["foo"],
        "description": "This is just a job for testing purposes.\n",
        "description_html": "<p>This is just a job for testing purposes.</p>\n",
        "parent": "cool-base-job",
        "url": "https://github/zuul.d/jobs.yaml",
        "private": False,
        "platforms": [],
        "scrape_time": scrape_time,
        "line_start": 1,
        "line_end": 7,
        "last_updated": None,
    }

    expected_job_2 = {
        "job_name": "another-job",
        "repo": "my/project",
        "tenants": ["foo"],
        "description": "This time without a playbook and a parent.\n",
        "description_html": "<p>This time without a playbook and a parent.</p>\n",
        "parent": "base",
        "url": "https://github/zuul.d/jobs.yaml",
        "private": False,
        "platforms": [],
        "scrape_time": scrape_time,
        "line_start": 8,
        "line_end": 12,
        "last_updated": None,
    }

    expected_job_3 = {
        "job_name": "cool-base-job",
        "repo": "my/project",
        "tenants": ["foo"],
        "description": "This is a base job with explicitly no parent.\n",
        "description_html": "<p>This is a base job with explicitly no parent.</p>\n",
        "parent": None,
        "url": "https://github/zuul.d/jobs.yaml",
        "private": False,
        "platforms": [],
        "line_start": 13,
        "line_end": 17,
        "scrape_time": scrape_time,
        "last_updated": None,
    }

    expected_role_1 = {
        "role_name": "foo",
        "repo": "my/project",
        "tenants": ["foo", "bar"],
        "description": "Just some simple description\n\n"
        ".. supported_os:: Linux, Windows\n",
        "description_html": "<p>Just some simple description</p>\n",
        "url": "https://github/my/project/tree/master/roles/foo",
        "private": False,
        "platforms": ["linux", "windows"],
        "scrape_time": scrape_time,
        "last_updated": "2018-09-17 15:15:15",
    }

    expected_role_2 = {
        "role_name": "bar",
        "repo": "my/project",
        "tenants": ["foo", "bar"],
        "description": "Role description containing some reStructuredText expressions.\n\n"
        "**Role variables**\n\n"
        ".. zuul:rolevar:: mandatory_variable\n\n"
        "   This variable is mandatory.\n\n\n"
        ".. zuul:rolevar:: optional_variable\n"
        "   :default: some_value\n\n"
        "   This one is not.\n\n\n"
        ".. zuul:rolevar:: list_variable\n"
        "   :default: []\n"
        "   :type: list\n\n"
        "   This one must be a list.\n",
        "description_html": "<p>Role description containing some reStructuredText "
        "expressions.</p>\n"
        "<p><strong>Role variables</strong></p>\n"
        '<dl class="rolevar">\n'
        '<dt id="rolevar-mandatory_variable">\n'
        '<code class="descname">mandatory_variable</code><a '
        'class="headerlink" href="#rolevar-mandatory_variable" '
        'title="Permalink to this definition">¶</a><br /></dt>\n'
        "<dd><p>This variable is mandatory.</p>\n"
        "</dd></dl>\n"
        "\n"
        '<dl class="rolevar">\n'
        '<dt id="rolevar-optional_variable">\n'
        '<code class="descname">optional_variable</code><a '
        'class="headerlink" href="#rolevar-optional_variable" '
        'title="Permalink to this definition">¶</a><br />Default: '
        '<code class="docutils literal notranslate"><span '
        'class="pre">some_value</span></code><br /></dt>\n'
        "<dd><p>This one is not.</p>\n"
        "</dd></dl>\n"
        "\n"
        '<dl class="rolevar">\n'
        '<dt id="rolevar-list_variable">\n'
        '<code class="descname">list_variable</code><a '
        'class="headerlink" href="#rolevar-list_variable" '
        'title="Permalink to this definition">¶</a><br />Default: '
        '<code class="docutils literal notranslate"><span '
        'class="pre">[]</span></code><br /></dt>\n'
        "<dd><p>This one must be a list.</p>\n"
        "</dd></dl>\n"
        "\n",
        "url": "https://github/my/project/tree/master/roles/bar",
        "private": False,
        "platforms": [],
        "scrape_time": scrape_time,
        "last_updated": "2018-09-17 15:15:15",
    }

    # NOTE (fschmidt): Without the skip_empty flag, empty (= None) keys will
    # be stripped from the resulting dict.
    assert job_1.to_dict(skip_empty=False) == expected_job_1
    assert job_2.to_dict(skip_empty=False) == expected_job_2
    assert job_3.to_dict(skip_empty=False) == expected_job_3
    assert role_1.to_dict(skip_empty=False) == expected_role_1
    assert role_2.to_dict(skip_empty=False) == expected_role_2



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from zubbi.utils import urljoin

LOGGER = logging.getLogger(__name__)


class GitConnection:
    def __init__(
        self, url, user=None, password=None, workspace="/tmp/zubbi_working_dir"
    ):
        self.git_host_url = url

        # TODO Support password via envvar (like in zapfel)
        self.user = user
        self.password = password

        # TODO (felix): Not sure if the connection is the right scope for this variable,
        # but it's the simplest way to configure it - and one could argue, that the
        # workspace directory is depending on the connection entry in the settings file
        # (different connection -> different workspace)
        self.workspace_dir = workspace

        # TODO If we want to support ssh and https, we should add a protocol parameter

    def init(self):
        LOGGER.info("Initializing Git connection to %s", self.git_host_url)
        # Currently, we don't need to do anything

    def get_remote_url(self, repository_name):
        # TODO (felix): Find a better way to rebuild the url with auth values
        # TODO (felix): Currently, this method requires the git_host_url to
        # provide a schema (e.g. https://)
        scheme, _, url = self.git_host_url.partition("://")

        if self.user and self.password:
            auth_part = "{}:{}@".format(self.user, self.password)
        elif self.user:
            auth_part = "{}@".format(self.user)
        else:
            auth_part = ""

        auth_base_url = "{}://{}{}".format(scheme, auth_part, url)
        remote_url = urljoin(auth_base_url, repository_name)
        return remote_url

    @property
    def provider(self):
        return "git"



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
import logging
import os
import pathlib
import sys
import tempfile

from docutils.parsers.rst import Directive
from readme_renderer import markdown
from sphinx.application import Sphinx
from sphinx.util import logging as sphinx_logging
from sphinx.util.console import nocolor
from sphinx.util.docutils import docutils_namespace, patch_docutils


SUPPORTED_OS_LOG_PREFIX = "supported-os"
LOGGER = logging.getLogger(__name__)


class SupportedOS(Directive):
    has_content = True
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True

    logger = sphinx_logging.getLogger("supported-os")

    def run(self):
        if len(self.arguments) > 0:
            platforms = self.arguments[0].split(",")
            for platform in platforms:
                # Log the platform info in a dedicated logger for later extraction
                self.logger.info("%s: %s", SUPPORTED_OS_LOG_PREFIX, platform)
        # We don't want to render anything, so we return an empty list of nodes
        return []


class SphinxBuildError(RuntimeError):
    pass


def render_sphinx(content):
    with tempfile.TemporaryDirectory() as tmp_dir:
        src_path = pathlib.Path(tmp_dir, "src/contents.rst")
        src_path.parent.mkdir()
        with src_path.open("w") as src:
            src.write(content)

        build_path = pathlib.Path(tmp_dir, "build/contents.fjson")

        source_dir = str(src_path.parent)
        doctree_dir = os.path.join(source_dir, ".doctrees")
        confoverrides = {"extensions": ["zuul_sphinx"]}
        status_log = io.StringIO()

        # NOTE (fschmidt): This part needs to be in sync with the used version
        # of Sphinx. Current version is:
        # https://github.com/sphinx-doc/sphinx/blob/v1.8.1/sphinx/cmd/build.py#L299
        with patch_docutils(source_dir), docutils_namespace():
            # Remove the color from the Sphinx' console output. Otherwise
            # the lines cannot be parsed properly as some \n are not set properly.
            nocolor()
            app = Sphinx(
                srcdir=source_dir,
                confdir=None,
                outdir=str(build_path.parent),
                doctreedir=doctree_dir,
                buildername="json",
                confoverrides=confoverrides,
                status=status_log,
                warning=sys.stderr,
            )

            # Add the mocked SupportedOS directive to get the os information
            # without rendering it into the resulting HTML page
            app.add_directive("supported_os", SupportedOS)

            # Start the Sphinx build
            app.build(force_all=True, filenames=[])

            if app.statuscode:
                raise SphinxBuildError

        # Extract the platforms from the logger output
        platforms = []
        status_log.seek(0)
        for line in status_log.readlines():
            prefix, _, platform = line.partition(":")
            if prefix == SUPPORTED_OS_LOG_PREFIX:
                platforms.append(platform.strip().lower())

        with build_path.open() as build:
            html_parts = json.load(build)

    return {"html": html_parts["body"], "platforms": platforms}


def render_markdown(content):
    # NOTE (fschmidt): We want to return a similar result like the parse_sphinx()
    # function.
    rendered = markdown.render(content)
    return {"html": rendered}


def render_file(file_dict):
    filepath = file_dict["path"]
    content = file_dict["content"]
    # Render the role description based on the file extension
    if filepath.lower().endswith(".rst"):
        LOGGER.debug("Rendering reStructuredText description")
        try:
            doc = render_sphinx(content)
            return doc
        except SphinxBuildError as exc:
            LOGGER.warning(
                "Content of %s could not be converted to HTML: %s", filepath, exc
            )
        except LookupError:
            LOGGER.exception(
                "Sphinx build failed. Most probably due to the usage of an invalid Sphinx directive or Zuul variable type."
            )
    elif filepath.lower().endswith(".md"):
        LOGGER.debug("Rendering markdown description")
        doc = render_markdown(content)
        return doc
    else:
        # Otherwise, we won't render the description at all.
        # In the UI we could use the missing description_html as
        # indicator to show the how-to-document link
        LOGGER.debug("Found txt or raw description. Skip rendering")



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from zubbi.scraper.exceptions import CheckoutError


LOGGER = logging.getLogger(__name__)

ZUUL_DIRECTORIES = ["zuul.d", ".zuul.d"]

ZUUL_FILES = ["zuul.yaml", ".zuul.yaml"]

README_FILES = ["README.rst", "README.md", "README.txt", "README"]

CHANGELOG_FILES = ["CHANGELOG.rst", "CHANGELOG.md", "CHANGELOG.txt", "CHANGELOG"]

REPO_ROOT = "/"


class Scraper:
    def __init__(self, repo):
        self.repo = repo

    def scrape(self):
        LOGGER.info("Scraping '%s'", self.repo.repo_name)

        job_files = self.check_out_job_files()
        role_files = self.check_out_role_files()

        return job_files, role_files

    def check_out_job_files(self):
        job_files = {}

        root_files = self.repo.list_directory(REPO_ROOT)

        # Search zuul directories
        for directory in ZUUL_DIRECTORIES:
            # Skip non-existing zuul dirs
            if directory not in root_files.keys():
                continue
            try:
                remote_files = self.repo.list_directory(directory)
                for filename, file_content in remote_files.items():
                    rel_path = file_content.path
                    try:
                        job_files[rel_path] = {
                            "last_changed": self.repo.last_changed(rel_path),
                            "blame": self.repo.blame(rel_path),
                            "content": self.repo.check_out_file(rel_path),
                        }
                    except CheckoutError as e:
                        LOGGER.exception(e)
            except CheckoutError as e:
                LOGGER.debug(e)

        # Search zuul files
        for zuul_file in ZUUL_FILES:
            # Skip non-existing zuul files
            if zuul_file not in root_files.keys():
                continue
            try:
                job_files[zuul_file] = {
                    "last_changed": self.repo.last_changed(zuul_file),
                    "blame": self.repo.blame(zuul_file),
                    "content": self.repo.check_out_file(zuul_file),
                }
            except CheckoutError as e:
                LOGGER.debug(e)

        return job_files

    def check_out_role_files(self):
        role_files = {}
        # Try to access the 'roles' directory
        try:
            roles = self.repo.list_directory("roles")
            for role_name, role_content in roles.items():
                try:
                    last_changed = self.repo.last_changed(role_content.path)
                    existing_files = self.repo.list_directory(role_content.path)
                    # Skip empty directories or files
                    if not existing_files:
                        continue
                    readme_file = self.find_matching_file(README_FILES, existing_files)
                    changelog_file = self.find_matching_file(
                        CHANGELOG_FILES, existing_files
                    )
                    role_files[role_name] = {
                        "last_changed": last_changed,
                        "readme_file": readme_file,
                        "changelog_file": changelog_file,
                    }
                except CheckoutError as e:
                    LOGGER.exception(e)
        except CheckoutError as e:
            LOGGER.debug(e)

        return role_files

    def find_matching_file(self, file_filter, existing_files):
        for filename, file_content in existing_files.items():
            if filename not in file_filter:
                continue
            try:
                rel_path = file_content.path
                # Return the first matching file
                match = {
                    "path": rel_path,
                    "content": self.repo.check_out_file(rel_path),
                }
                return match
            except CheckoutError as e:
                LOGGER.exception(e)
